/*
 *	Copyright (c) Vanderbilt University, 2001-2002
 *	ALL RIGHTS RESERVED
 *
 *	Vanderbilt University disclaims all warranties with regard to this
 *	software, including all implied warranties of merchantability
 *	and fitness.  In no event shall Vanderbilt University be liable for
 *	any special, indirect or consequential damages or any damages
 *	whatsoever resulting from loss of use, data or profits, whether
 *	in an action of contract, negligence or other tortious action,
 *	arising out of or in connection with the use or performance of
 *	this software.
 *
 *  FILE NAME:	Effector.cpp 
 *	SYNOPSIS:	Effector of Graph Rewrite Engine Implementation
 *  AUTHOR:		Feng Shi - feng.shi@vanderbilt.edu
 *  CREATED:	 11/13/02
 *  LAST UPDATE: 
 */

#pragma warning ( disable : 4786 ) // Names more than 255 characters

#include <Windows.h>
#include "sstream"
extern HINSTANCE hCINTDLL;

#include "Effector.h"

#include "GReATUtils.h"

Effector::Effector(GR::RewritingRule *currRule,GeneralBackend *dataNetworkMap) :
	boundRule(currRule),
	backendMap(dataNetworkMap)
{
}

Effector::~Effector() {}

void Effector::executeRightSide(Match *mat)
{	
//Execute order:
//	- binding
//	- CreateNew
//	- AttributeMapping
//	- Delete
	match = mat;
	bindingConsequenceObject();   //Composition

// Need check all Consequence Links, to make sure that 
// all associations have been created except the composition
	bindingRemainingObjectLinks(); 
	executeCode();
	reqDelete();
}

void Effector::bindingConsequenceObject()
{
// This function need to be called recursively
	GR::ObjectWrapper parentCO;
	GR::ObjectWrapper childCO;
	
	Udm::Object::CompositionInfo compInfo;
	 
	if(getCompositionToBound(parentCO,childCO, compInfo))
	{
		string pname = parentCO.name();
		string cname = childCO.name();
		Object outputChild = backendMap->CreateUdmObject(parentCO,childCO,compInfo);
		if(outputChild != Udm::null)
		{	
			childCO.objectID() = outputChild.uniqueId();
			childCO.dataNetworkID() = outputChild.__impl()->__getdn()->uniqueId();
			match->grObj2BoundObj[childCO] = outputChild;
			match->boundObject.insert(pair<unsigned long, Object>(childCO.dataNetworkID(), outputChild));
		}	 

		bindingConsequenceObject();
	}
/*	else 
	{	
		const set<GR::PatternObject>& pObjs = boundRule->PatternObject_kind_children();
//First, remove all association class with action "Delete"		
		for(set<GR::PatternObject>::const_iterator i=pObjs.begin();i!=pObjs.end();++i)
		{
			GR::PatternObject pObj = *i;
			if(!pObj.isAssociationClass()) continue;
			if((string)pObj.action()=="Delete")	
			{
				backendMap->DeleteUdmObject(pObj);
			}
		}
//Second, remove all other objects with action "Delete"
		for(set<GR::PatternObject>::const_iterator j=pObjs.begin();j!=pObjs.end();++j)
		{
			GR::PatternObject pObj = *j;
			if(pObj.isAssociationClass()) continue;
			if((string)pObj.action()=="Delete")	
			{
				backendMap->DeleteUdmObject(pObj);
			}
		}
	}
*/
}

void Effector::reqDelete()
{
	const set<GR::PatternObject>& pObjs = boundRule->PatternObject_kind_children();
//First, remove all association class with action "Delete"		
	for(set<GR::PatternObject>::const_iterator i=pObjs.begin();i!=pObjs.end();++i)
	{
		GR::PatternObject pObj = *i;
		if(!pObj.isAssociationClass()) continue;
		if((string)pObj.action()=="Delete")	
		{
			backendMap->DeleteUdmObject(pObj);
		}
	}
//Second, remove all other objects with action "Delete"
	for(set<GR::PatternObject>::const_iterator j=pObjs.begin();j!=pObjs.end();++j)
	{
		GR::PatternObject pObj = *j;
		if(pObj.isAssociationClass()) continue;
		if((string)pObj.action()=="Delete")	
		{
			backendMap->DeleteUdmObject(pObj);
		}
	}
}

bool Effector::getCompositionToBound(GR::ObjectWrapper &parentCO,GR::ObjectWrapper &childCO, Udm::Object::CompositionInfo &compInfo)
{
	const set<GR::ObjectLink>& consequenceLinks=boundRule->ObjectLink_kind_children();

	for(set<GR::ObjectLink>::const_iterator i=consequenceLinks.begin();\
										   i!=consequenceLinks.end(); ++i)
	{
		const GR::ObjectLink& clink = *i;
		if((string)clink.relationshipType()!=(string)"Composition" || (string)clink.cardinality()=="0") continue;

		const GR::ObjectWrapper& dstCO = clink.dst_end();
		const GR::ObjectWrapper& srcCO = clink.src_end();

		ASSERT((string)dstCO.type().name()==(string)srcCO.type().name());

		//1-2-04
		//Bug found here, consider the issue that the composition maybe "0"
		if( dstCO.objectID()!=0 && srcCO.objectID()==0)
		{
			if((string)srcCO.action()!="CreateNew" && (string)clink.action()=="CreateNew")
				throw udm_exception("Pattern object \""+(string)srcCO.name()+"\" in rule \""+(string)boundRule->name()+"\" should have action \"CreateNew\".");
	
			parentCO=dstCO;
			childCO=srcCO;
			compInfo.strParentRoleName = clink.dstRoleName();
			compInfo.strChildRoleName = clink.srcRoleName();
			return true;
		}
	}

	return false;
}

void Effector::bindingRemainingObjectLinks()
{
	const set<GR::ObjectLink>& consequenceLinks=boundRule->ObjectLink_kind_children();

	for(set<GR::ObjectLink>::const_iterator i=consequenceLinks.begin();\
										   i!=consequenceLinks.end(); ++i)
	{
		GR::ObjectLink clink = *i;
		if((string)clink.action()=="CreateNew" && (string)clink.relationshipType()!="Composition")
		{
			backendMap->CreateUdmAssociation(clink);
		}
		if((string)clink.action()=="Delete" && (string)clink.relationshipType()!="AssociationClass") 
		{
			GR::ObjectWrapper srcCO = clink.src_end();
			GR::ObjectWrapper dstCO = clink.dst_end();

			if((string)srcCO.action()=="Delete" || (string)dstCO.action()=="Delete")
				continue;

			backendMap->DeleteUdmAssociation(clink);
		}
	}
}

typedef void (*MAP_FN)(const vector <Udm::Object> &objs, bool& err);
void Effector::executeCode()
{	
	const vector<GR::Code> &ams = boundRule->Code_kind_children();
	if(ams.size()==0)
		return;

	ostringstream gFStream;
	const __int64 &o_id = boundRule->ruleId();
	gFStream << "AM" << "_" << convertToCPlusPlusName((string)boundRule->name()) << "_" << o_id ;
	string fn = gFStream.str();

	//cout << "Executing AM function " << fn << endl;

  MAP_FN lpMapFn = (MAP_FN)GetProcAddress(hCINTDLL, fn.c_str());

	if(lpMapFn == NULL)
	{	
    cout << "ERROR: AM function " << fn << " not found in dll \n";
		return;
	}

	vector <Udm::Object > objs;
	set<string > names;

	typedef NameCompareFunctor< GR::PatternObject> PatternObjectCompareFunc;
	typedef set< GR::PatternObject, PatternObjectCompareFunc> PattObjSorted_t;
	PattObjSorted_t patobjs = boundRule->PatternObject_kind_children_sorted( PatternObjectCompareFunc());
	PattObjSorted_t::const_iterator i;
//	for(i=patobjs.begin();i!=patobjs.end();i++)
//	const vector<GR::PatternObject> &patobjs = boundRule->PatternObject_kind_children();
//	vector<GR::PatternObject>::const_iterator i;
	for(i=patobjs.begin();i!=patobjs.end();i++)
	{	const GR::PatternObject &po = *i;
		string name = po.name();
		set<string>::iterator ni = names.find(name);
		if(ni!=names.end())
			continue;
		names.insert(name);
		unsigned long o_id = (unsigned long)po.objectID();
//		Udm::Object o = match->grObj2BoundObj[po]; //Bug here
//		objs.push_back(o);
		Udm::Object o;
		if(match->grObj2BoundObj.find(po)!=match->grObj2BoundObj.end())
		{
			o = match->grObj2BoundObj[po];
		}
		objs.push_back(o);
	}

  bool err = false;
  lpMapFn(objs, err);
  if (err)
    cout << "ERROR: AM function " << fn << " could not execute \n";
	return;
}