#include "PacketPassing.h"
#include "RuleTypes.h"
#include "PEStringTable.h"
#include "GReATUtils.h"


// This needed for insert_iterator
// In case of std::set-s, new elements are stored in a sorted sequence.
bool operator<( const ParamData& first, const ParamData& second)
{
	return ( first._param < second._param);
}

bool operator==( const ParamData& first, const ParamData& second)
{
	return ( first._param== second._param);
}

///////////////////////////////////////////////////////////////////////////////
// ParamList
void ParamList::operator () ( const GR::ObjectWrapper& ow)
{
	if ( isInput()( ow))
	{
		const std::string& paramName= addInputParam( ow, *_function);
		// Container independent insert (just for containers with insert() method, so semi-useful stuff)
		*insert_iterator<ParamDatas_t>( _inputParams, _inputParams.end())= ParamData( ow, paramName);
	}
	if ( isOutput()( ow))
	{
		const std::string& paramName= addOutputParam( ow, *_function);
		// Container independent insert (just for containers with insert() method, so semi-useful stuff)
		*insert_iterator<ParamDatas_t>( _outputParams, _outputParams.end())= ParamData( ow, paramName);
	}
}

const std::string& ParamList::addInputParam( const GR::ObjectWrapper& ow, Function& function) const
{
	FuncParam_t inputParam= Function::createInputParam( lowerBegin( ( std::string) ow.name()+ "s"));
	const FuncParam_t& inserted= function.insertInputParam( inputParam);
	return inserted._param.second;
}

const std::string& ParamList::addOutputParam( const GR::ObjectWrapper& ow, Function& function) const
{
	FuncParam_t outputParam= Function::createOutputParam( lowerBegin(( std::string) ow.name()+ "s"));
	const FuncParam_t& inserted= function.insertOutputParam( outputParam);
	return inserted._param.second;
}

///////////////////////////////////////////////////////////////////////////////
// non-member helper functions
void registerPort( const GR::ObjectWrapper& port, const ClassDef* memberDef, ParamDatas_t& params) throw ( CGEx)
{
	ParamDatas_t::iterator itFindParam= std::find( params.begin(), params.end(), ParamData( port, ""));
	if ( params.end() == itFindParam)
		throw CGEx("FATAL: Trying to register port belonging to a non-existing parameter.");
	itFindParam->_memberDef= memberDef;
}

const ClassDef* getPort( const GR::ObjectWrapper& port, const ParamDatas_t& params)
{
	ParamDatas_t::const_iterator itFindParam= std::find( params.begin(), params.end(), ParamData( port, ""));
	if ( params.end() == itFindParam)
		return NULL;
	return itFindParam->_memberDef;
}

///////////////////////////////////////////////////////////////////////////////
// PacketPassing
bool PacketPassing::isAnyDeletedObjectAmongParams( const ParamDatas_t& params)
{
	ParamDatas_t::const_iterator itFind= find_if( params.begin(), params.end(), isDeleted());
	return itFind != params.end();
}

bool PacketPassing::allOutputsAreInputs( const ParamDatas_t& outputs)
{
	ParamDatas_t::const_iterator itFind= find_if( outputs.begin(), outputs.end(), not1( isInput()));
	return ( itFind == outputs.end());
}

void PacketPassing::createParamList( const GR::RewritingRule& rule, Function& function)
{
	std::set< GR::ObjectWrapper> ows= rule.ObjectWrapper_kind_children();
	ParamList params( function);
	for_each_ref( ows.begin(), ows.end(), params);
	_ruleParams[ rule]= params;
}

const std::string& PacketPassing::addInputIterator( const ParamData& inputParam, Loop& inputLoop) const
{
	std::string iteratorName= PE::createUniqueName( IT+ ( std::string)inputParam._param.name());
	inputLoop.addInputIterator( iteratorName, inputParam._name);
	return inputLoop._loopVariable;
}

void PacketPassing::matchParams( const GR::RuleBase& source, const GR::RuleBase& destination, std::list< std::string>& callerParams, bool findSourceInInputs) const
{
	// Parameter matching runs on parameters of the generated class function operators.
	// If source or destination is a proxy, the referred RewritingRule is used for sequencing 
	// input/ output parameters, and parameter name matching. But pass along tracking is
	// performed on object wrapper proxies of the rule proxies.
	const ParamDatas_t* inputs= inputParams( getRewritingRule( destination));
	if ( NULL== inputs)
		throw CGEx("FATAL: Inputs of destination rule: "+ ( std::string)destination.name()+ " not found while matching call parameters.");
	const ParamDatas_t* sources= ( findSourceInInputs) ? inputParams( getRewritingRule( source)) : outputParams( getRewritingRule( source));
	if ( NULL== sources)
		throw CGEx("FATAL: Inputs/outputs of source rule: "+ ( std::string)source.name()+ " not found while matching call parameters.");
	// Iterate inputs of func.op. to call.
	for( ParamDatas_t::const_iterator itInputs= inputs->begin(); itInputs!= inputs->end(); ++itInputs)
	{
		std::list< GR::ObjectWrapper> sourceOWs;	// Contains the source object wrappers of the current input.
		GR::ObjectWrapper currInputOW= itInputs->_param;
		// Check if currInputOW is a referred object wrapper by any of source's object wrappers. 
		// Update 'currInputOW' with the referer, if neccessary.
		currInputOW= getInputOWForPATracking( destination, currInputOW);
		// Get source pass alongs of 'currInputOW'.
		std::set< GR::PassAlong> srcPAs= currInputOW.srcObjectWrapper();
		for ( std::set< GR::PassAlong>::const_iterator itSrcPAs= srcPAs.begin(); itSrcPAs!= srcPAs.end(); ++itSrcPAs)
		{
			const GR::ObjectWrapper currSrcOW= itSrcPAs->srcObjectWrapper_end();
			if ( GR::RuleBase::Cast( currSrcOW.parent()) == source)
				sourceOWs.push_back( currSrcOW);
		}
		// there must be only one source for every input!
		if ( 1!= sourceOWs.size())
			throw CGEx("FATAL: Input: "+ ( std::string)currInputOW.name()+ " in rule: "+ ( std::string)destination.name()+ " has other than one source while matching call parameters.");
		GR::ObjectWrapper sourceOW= sourceOWs.front();
		// If 'source' is of type 'GR::RewritingRuleProxy', then the found 'sourceOW' is of type 'GR::ObjectWrapperProxy'.
		// In that case, the parameter name is gained by searching 'sourceOW.refersTo()' in 'sources', because
		// only the referred rule has a parameter std::list defined.
		const ParamData& findSource= findParamData( *sources, sourceOW);
		callerParams.push_back( findSource._name);
	}
}

const ParamDatas_t* PacketPassing::inputParams( const GR::RewritingRule& rule) const
{
	RuleParamMap_t::const_iterator itFind= _ruleParams.find( rule);
	return ( itFind== _ruleParams.end()) ? NULL : ( itFind->second.inputParams());
}

const ParamDatas_t* PacketPassing::outputParams( const GR::RewritingRule& rule) const
{
	RuleParamMap_t::const_iterator itFind= _ruleParams.find( rule);
	return ( itFind== _ruleParams.end()) ? NULL : ( itFind->second.outputParams());
}

ParamDatas_t* PacketPassing::inputParams( const GR::RewritingRule& rule)
{
	RuleParamMap_t::iterator itFind= _ruleParams.find( rule);
	return ( itFind== _ruleParams.end()) ? NULL : ( itFind->second.inputParams());
}

ParamDatas_t* PacketPassing::outputParams( const GR::RewritingRule& rule)
{
	RuleParamMap_t::iterator itFind= _ruleParams.find( rule);
	return ( itFind== _ruleParams.end()) ? NULL : ( itFind->second.outputParams());
}

GR::ObjectWrapper PacketPassing::getInputOWForPATracking( const GR::RuleBase& ruleBase, const GR::ObjectWrapper ow) const
{
	if( true== Uml::IsDerivedFrom( ruleBase.type(), GR::RewritingRuleProxy::meta))
	{
		std::set< GR::ObjectWrapperProxy> owProxies= GR::RewritingRuleProxy::Cast( ruleBase).ObjectWrapperProxy_children();
		for( std::set< GR::ObjectWrapperProxy>::const_iterator itOWProxies= owProxies.begin(); itOWProxies!= owProxies.end(); ++itOWProxies)
		{
			if (( GR::ObjectWrapper::Cast( itOWProxies->refersTo())== ow) && ( true== isInput()( *itOWProxies)))
			{
				return *itOWProxies;
			}
		}
		throw CGEx("FATAL: Object wrapper proxy referring to object wrapper: "+ ( std::string)ow.name()+  "not found in rewriting rule proxy: "+ ( std::string)ruleBase.name());
	}
	return ow;
}

const ParamData& PacketPassing::findParamData( const ParamDatas_t& params, const GR::ObjectWrapper& owToFind) const
{
	ParamData findSourceOW( owToFind, "");
	if ( true== Uml::IsDerivedFrom( owToFind.type(), GR::ObjectWrapperProxy::meta))
		findSourceOW._param= ( GR::ObjectWrapperProxy::Cast( owToFind)).refersTo();
	ParamDatas_t::const_iterator itFindSrc= find( params.begin(), params.end(), findSourceOW);
	if ( itFindSrc== params.end())
		throw CGEx("FATAL: Source object wrapper: "+ ( std::string)owToFind.name()+ " not found in sources.");
	return *itFindSrc;
}
