/* NAME:	Translator
 * AUTHOR:	Attila Vizhanyo - viza@isis-server.isis.vanderbilt.edu
 *
 * Translates GR data structure to PE data structure
 */
#ifndef __TRANSLATOR_H
#define	__TRANSLATOR_H

#ifdef _MSC_VER
#pragma warning (disable : 4786)
#pragma warning (disable : 4503) 
#endif

#include "RuleTypes.h"
#include "PacketPassing.h"
#include "PE.h"
#include "GeneralBackend.h"
#include "ICIntCompiler.h"

// Translates GR rules and keeps track of <GR rule - generated PE element> pairs.
// This class is not designed to be a base class. 
// Either don't inherit from it, or add virtual destructor.
class Translator
{
public:
	// ctor; see parameter desc. at the member variables declaration part. 
	Translator( const GeneralBackend& backend, PacketPassing& passPacket, Program& program, ICIntCompiler& cc) : _backend( backend), _passPacket( passPacket), _program( program), _cc( cc) {}
	// Translate different types of 'GR::RewritingRule'-s (see: "RuleTypes.h") to their PE class definition.
	void translateBlock( const GRBlock& block);
	void translateForBlock( const GRForBlock& block);
	void translateRule( const GRRule& rule);
	void translateTest( const GRTest& test);
	void translateForTest( const GRForTest& forTest);
	void translateCase( const GRCase& testCase);
	// Translate case call of 'testCase' into test PE class definition: 'test'.
	void translateCallInTest( const GRTestType& test, const GRCase& testCase);	
	// If destination of 'seq' is output port, generate code which appends all output ports of block with source of 'seq'.
	// If destination of 'seq' is rule, generate (or use the existing, if there is) caller function which makes the call.
	// Add caller function into block class definiton.
	// Block is considered to be 'seq.parent()'.
	void translateCallInBlock( const GR::Sequence& seq);
	// Rule definition outlet
	const Class* ruleDef( const GR::RewritingRule& rule) const
	{
		return getRuleDef( rule);
	}
	//
	void updateCIntCompiler( ICIntCompiler& cc) { _cc= cc;}

protected:
	// Taking 'rule', creates and returns PE class definition: 'ruleDef' and func.operator: 'funcOperator'.
	void translateRuleDef( const GR::RewritingRule& rule, Class** ruleDef, Function** funcOperator);
	// Append 'ruleDef' with output definitions of outputs of 'rule', and initialize them in 'initFunc'.
	void translateOutputs( const GR::RewritingRule& rule, Class* ruleDef, Function* initFunc);
	// Append 'ruleDef' with the 'executeOne()' protected member function, and create the appropriate call into 'initFunc'.
	void translateForCall( const GR::RewritingRule& forRule, Class* ruleDef, Function* initFunc);
	// Append 'ruleDef' with the 'patternMatcher()' protected member function, and create the appropriate call into 'patternFuncCaller'.
	// Returns with pointer to the generated function definition in 'pipFuncDef', the loop traversing inputs in 'inputLoop'.
	// 'inputBoundObjects' is optimal parameter; if not null, then itParamNames for all inputs will be filled into it.
	void translatePatterns( const GRSimpleRule& rule, Class* ruleDef, PEComposite* patternFuncCaller, std::map< GR::ObjectWrapper, std::string>& boundObjects, Function* &pipFuncDef, Loop* &inputLoop, std::map< GR::ObjectWrapper, std::string>* inputBoundObjects= 0);
	// Append 'ruleDef' with the 'forwardInputs()' protected member function, and create the appropriate call into 'forwardFuncCaller'.
	// 'OutputAppender' is pointer to the OutputAppender() function definition.
	// Returns with pointer to the generated function definition.
	Function* translateForwardInputs( const GRRule& rule, Class* ruleDef, PEComposite* forwardFuncCaller, Function* outputAppenderFunc);
	// Append 'ruleDef' with the 'effector()' protected member function, and create the appropriate call into 'effectorFuncCaller'.
	// Returns with a pointer to the generated match-loop body of 'effector()'.
	PEComposite* translateEffector( const GRRule& rule, Class* ruleDef, PEComposite* effectorFuncCaller, std::map< GR::ObjectWrapper, std::string>& boundObjects);
	// Append 'ruleDef' with the 'outputAppender()' protected member function, and create the appropriate call into 'outputAppenderFuncCaller'.
	// Returns with pointer to the generated function definition.
	Function* translateOutputAppender( const GRSimpleRule& rule, Class* ruleDef, PEComposite* outputAppenderFuncCaller, std::map< GR::ObjectWrapper, std::string>& boundObjects);
	// Add append-output calls for each source in 'sources' to 'caller', which has destination in 'outputParent'.
	void addAppendOutputCalls( const ParamDatas_t& sources, const GR::RewritingRule& outputParent, Function& caller) const;
	// Returns with pointer to 'operator()'/ 'executeOne()' for blocks/ for blocks.
	Function* getBlockCaller( const GRBlockType& block) const;
	// Returns with pointer to 'operator()'/ 'executeOne()' for tests/ for tests.
	Function* getTestCaller( const GRTestType& test) const;
	// Caller function is a protected member of class generated for 'block', 
	// which makes a func.op.call on the class generated for 'toCall'.
	// Creates caller function (if not existed), and returns with pointer to this function.
	Function* getOrCreateCaller( const GR::RuleBase& toCall, const GRBlockType& block);
	// Rule definition maintenance.
	void registerRuleDef( const GR::RewritingRule& rule, const Class* ruleDef);
	Class* getRuleDef( const GR::RewritingRule& rule) const;
	// Caller function maintenance.
	void registerCaller( const GR::RuleBase& ruleBase, const Function* caller);
	Function* getCaller( const GR::RuleBase& ruleBase) const;
	// Returns true, if rule base 'rule' is isolated, that is, it has no sources.
	bool isIsolated( const GR::RuleBase& rule) const;
	// Returns true, if rule 'rule' is refered by any non-isolated proxies.
	bool isReferred( const GR::RewritingRule& rule) const;

private:
	// Maps rules to their definition class.
	typedef std::map< GR::RewritingRule, const Class*> RuleDefMap_t;
	RuleDefMap_t _ruleDefMap;
	// Maps rule bases, to their caller function counterparts.
	typedef std::map< GR::RuleBase, const Function*> CallerMap_t;
	CallerMap_t _callerMap;
	// General backend provides host meta information.
	const GeneralBackend& _backend;
	// Packet passing used for exploring packet passing bindings among rules. (See PacketPassing.h)
	PacketPassing& _passPacket;
	// Program output.
	Program& _program;
	//
public:
	ICIntCompiler& _cc;
};

#endif //__TRANSLATOR_H
