/* NAME:	Smart DN Code Printer
 * AUTHOR:	Attila Vizhanyo - viza@isis.vanderbilt.edu
 *
 * Code Printer classes for smart data networks
 */
#ifndef SMARTDNCODEPRINTER_H
#define SMARTDNCODEPRINTER_H

#include "DNCodePrinter.h"

class SmartDNCodePrinter : public DNCodePrinter
{
public:
	SmartDNCodePrinter();
	virtual void generate( 
			const std::string& hostFileName,
			const std::string& copyFileName,
			const std::string& meta, 
			const std::string& root,
			const std::string& dtd
		)
	{
		generateOpen( hostFileName, meta, root, dtd);
		generateClose( copyFileName, meta, root, dtd);
	}
	virtual void generateOpen( const std::string& fname, const std::string& meta, const std::string& root, const std::string& dtd) = 0;
	virtual void generateClose( const std::string& fname, const std::string& meta, const std::string& root, const std::string& dtd) = 0;
	virtual void getOpenCode( PEComposite& addTo) const;
	virtual void getCloseCode( PEComposite& addTo) const;
	virtual void getDnCode( PEComposite& addTo) const;
	virtual const std::string& getDN() const
	{
		return _memDef->_variable;
	}

protected:
	void createOpenCode( const std::string& fname, const std::string& meta, const std::string& root, const std::string& dtd);
	void createOpenCodeWriteOnly( const std::string& meta, const std::string& root);
	void createCopyMemToHostDN();
	void createCopyMemToDN( const std::string& dn);
	void createCloseDNs( bool closeHostWithUpdate, bool closeMemWithUpdate);
	void createCloseHost( bool closeHostWithUpdate);
	void createCloseMem( bool closeMemWithUpdate);
	bool isDTDMeta( const std::string& fname) const;

private:
	Definition* _hostDef;
	Expression* _hostOpenExpr;
	Definition* _memDef;
	Expression* _memCreateExpr;
	Expression* _copyDNExpr;
	Expression* _copyBackDNExpr;
	Expression* _hostClose;
	Expression* _memClose;
};

class SmartDNReadOnlyCodePrinter : public SmartDNCodePrinter
{
public:
	virtual void generateOpen( const std::string& fname, const std::string& meta, const std::string& root, const std::string& dtd)
	{
		createOpenCode( fname, meta, root, dtd);
	}
	virtual void generateClose( const std::string& fname, const std::string& meta, const std::string& root, const std::string& dtd)
	{
		createCloseDNs( false, false);
	}
	virtual void getCloseCode( PEComposite& addTo) const
	{
		SmartDNCodePrinter::getCloseCode( addTo);
	}
};

class SmartDNWriteOnlyCodePrinter : public SmartDNCodePrinter
{
public:
	SmartDNWriteOnlyCodePrinter();
	virtual void generateOpen( const std::string& fname, const std::string& meta, const std::string& root, const std::string& dtd)
	{
		createOpenCodeWriteOnly( meta, root);
	}
	virtual void generateClose( const std::string& fname, const std::string& meta, const std::string& root, const std::string& dtd)
	{
		createNewDN( fname, meta, root, dtd);
		createCopyMemToNewDN();
		createCloseMem( true);
		createCloseNewDN();
	}
	virtual void getCloseCode( PEComposite& addTo) const
	{
		if ( _newHostDef) addTo.add( *_newHostDef);
		if ( _createNewHostExpr) addTo.add( *_createNewHostExpr);
		SmartDNCodePrinter::getCloseCode( addTo);
		if ( _closeNewHostExpr) addTo.add( *_closeNewHostExpr);
	}

protected:
	virtual const std::string getNewHostPrefix() 
	{
		return "sdnNew";
	}
	void createNewDN( const std::string& fname, const std::string& meta, const std::string& root, const std::string& dtd);
	void createCopyMemToNewDN();
	void createCloseNewDN();

private:
	Definition* _newHostDef;
	Expression* _createNewHostExpr;
	Expression* _closeNewHostExpr;
};

class SmartDNReadWriteCodePrinter : public SmartDNCodePrinter
{
public:
	virtual void generateOpen( const std::string& fname, const std::string& meta, const std::string& root, const std::string& dtd)
	{
		createOpenCode( fname, meta, root, dtd);
	}
	virtual void generateClose( const std::string& fname, const std::string& meta, const std::string& root, const std::string& dtd)
	{
		createCopyMemToHostDN();
		createCloseDNs( true, true);
	}
	virtual void getCloseCode( PEComposite& addTo) const
	{
		SmartDNCodePrinter::getCloseCode( addTo);
	}
};

class SmartDNWriteCopyCodePrinter : public SmartDNWriteOnlyCodePrinter
{
public:
	SmartDNWriteCopyCodePrinter() : SmartDNWriteOnlyCodePrinter() {}
	virtual void generateOpen( const std::string& fname, const std::string& meta, const std::string& root, const std::string& dtd)
	{
		createOpenCode( fname, meta, root, dtd);
	}
	virtual void generateClose( const std::string& fname, const std::string& meta, const std::string& root, const std::string& dtd)
	{
		SmartDNWriteOnlyCodePrinter::generateClose( fname, meta, root, dtd);
		createCloseHost( false);
	}
	virtual void getCloseCode( PEComposite& addTo) const
	{
		SmartDNWriteOnlyCodePrinter::getCloseCode( addTo);
	}

protected:
	virtual const std::string getNewHostPrefix()
	{
		return "sdnCopy";
	}
};

#endif SMARTDNCODEPRINTER_H