#if !defined(GRDebugInterface_h)
#define GRDebugInterface_h


#include "GRDebugTypes.h"
#include "GRDebugError.h"
#include "GRDebugEvent.h"

/// GR Debug namespace

namespace GRDI
{


/// Debug interface
class DebugInterface 
{
public:

///@name Execution Control
//@{

  ///Get status information of execution
  /**
  @param status on return, the current status of the execution. 
  The status is one of the values of @ref Status. 
  This function returns a @ref UniversalErrors 
  */
  virtual Error  getStatus(Status& status);

  ///Suspend the execution. 
  /**
    This function returns either a @ref UniversalErrors  or one of the following errors: 
    @ref GRDI_ERROR_ALREADY_SUSPENDED 
    Thread already suspended. 
  */
  virtual Error  suspend();

  ///Pause the execution. 
  virtual Error  pause();

  ///Resume a suspended execution
  /**
  This function returns either a @ref UniversalErrors or one of the following errors: 
  @ref GRDI_ERROR_NOT_SUSPENDED 
  Thread was not suspended. 
  */
  virtual Error  resume();

  ///Start the execution
  /**
   This function is used to start the executiom
  */
  virtual Error execute();

  ///Run
  /**
   This function is used to run
  */
  virtual Error run();

  ///Stop the execution
  /**
   This function is used to kill the executiom
  */
  virtual Error stop();


  ///Save a program
  /**
   This function saves a program.
  */
  virtual Error save();

  ///Load a program
  /**
   This function loads a program.
  */
  virtual Error load(const std::string&);
  ///Close a program
  /**
   This function closes program.
  */
  virtual Error close();

  ///Quit the seesion
  /**
   This function is used to kill gr engine
  */
  virtual Error quit();

//@}

///@name Context Stack Access
//@{

	/// Get the number of functions currently in the context stack. 
  /**
	@param count on return, the number of contextes in the context stack
	This function returns @ref UniversalErrors
  */

	virtual Error getContextStacktSize(int& count);


	/// Get the current context id
  /**
	@param contextID	on return, the contextID for the current stack postion. 
	This function returns either a @ref UniversalErrors or 
  @ref GRDI_ERROR_NO_MORE_CONTEXTES
  */

	virtual Error getCurrentContextID(ContextID& contextID);

	/// Get the value for the contextDescr
  /**
	@param contextDescr	on return, the contextDescr for the stack postion. 
	This function returns either a @ref UniversalErrors or 
  @ref GRDI_ERROR_NO_MORE_CONTEXTES
  */
  virtual Error getContextDescr(const ContextID& contextID, ContextDescr& contextDescr);

	/// For a context, return the context that called. 
  /**
	@param called the called context
	@param caller on return, the caller context  
	This function returns either a @ref UniversalErrors  or one of the following errors:
	@ref GRDI_ERROR_INVALID_CONTEXT
	@ref GRDI_ERROR_NO_MORE_CONTEXTES
  */

	virtual Error getCallerContext( const   ContextID& called,
								                  ContextID& caller);


	/// For a context, return the function currently executing 
	/**
  @param context the context to query. 
	@param func on return, the current function. 
	This function returns either a @ref UniversalErrors  or one of the following errors:
	@ref GRDI_ERROR_INVALID_CONTEXT 
  */

	virtual Error getCurrentLocation( const  ContextID& context,
									                  LocationID& locID);

 	/// Returns the description of the last exception.
  virtual Error getLastException( std::string& str);

 	/// Notify when context is popped.
  /**
  When function is popped from the stack, generate a GRDI_EVENT_CONTEXT_POP event. See @ref HandlingEvents. 
	@param context the context for which the context pop event will be generated. 
	This function returns either a @ref UniversalErrors  or one of the following errors:
	@ref GRDI_ERROR_INVALID_FUNCTION 
   */

  virtual Error notifyContextPop(const ContextID& context); 

//@}

///@name Breakpoints
//@{

	/// Set a breakpoint at the instruction indicated function and location. 
	/**
  A function location can only have one breakpoint. 
	@param func the function in which to set the breakpoint 
  @loc func the location on which to set the breakpoint within the function
  Whenever the designated function is about to be executed, 
  a @ref EVENT_BREAKPOINT event is generated. See @ref HandlingEvents. 
	This function returns either a @ref UniversalErrors  or one of the following errors:
	@ref GRDI_ERROR_INVALID_FUNCTION  
  @ref GRDI_ERROR_DUPLICATE 

  */
  virtual Error setBreakpoint(const LocationID& loc);


 	/// Clear the breakpoint at location indicated by function. 
  /**
	@param func the function in which to clear the breakpoint 
	This function returns either a @ref UniversalErrors  or one of the following errors:
	@ref GRDI_ERROR_INVALID_FUNCTION  
  @ref GRDI_ERROR_NOT_FOUND 

  */
  virtual Error clearBreakpoint(const LocationID& loc);


  /// Clear all breakpoints. 
  /// This function returns @ref UniversalErrors
  virtual Error  clearAllBreakpoints();

  ///
  virtual Error  getBreakPoints(LocationIDSet& breakpoints);
//@}

///@name IFunctionPtr locations 
//@{

  ///Gets start function 
//  virtual  Error getStartFunction(IFunctionPtr& startFunc);

  ///Gets root functions 
  virtual  Error getRootLocations(LocationIDVec& rootLocs);

  ///Gets functions
//  virtual  Error getFunctions(  FunctionDescrMap& funcDescrMap);
   
  
  ///Gets function's description

   virtual Error getFunctionDescr(const IFunctionPtr& func,
                                       FunctionDescr& descr);
 
  ///Gets function's locationID
  virtual Error getLocationFromID(const LocationID& locationID,
                                        Location& location);

///@name Events
//@{

  ///Set Event Hook
  /**
  Set the function to be called on every event. 
  Details on events are described in @ref HandlingEvents. 
  @param hook the new event hook, or NULL to remove the existing hook. 
  This function returns @ref UniversalErrors 
  */
  virtual Error setEventHook(EventHook* hook);



//@}

};
};
#endif //DebugInterface_h